#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "parser.h"
#include "ast.h"
#include "token.h"
#include "tokeniser.h"
#include "errors.h"
#include "dyn_arr.h"
#include "logger.h"

DEFINE_DA_IMPL(ast_node, void*);
DEFINE_DA_IMPL(func_param, ast_func_param_t);

token_t* expect_token(mcl_da_token_t* tks, token_t t);
token_t* try_token(mcl_da_token_t* tks, token_t t);
token_t* expect_token_err(mcl_da_token_t* tks, token_t t);
ast_func_decl_t* parse_func(mcl_da_token_t* tokens, ast_ident_t name);


ast_program_t* parse(mcl_da_token_t* tokens) {
    ast_program_t* prog = malloc(sizeof(ast_program_t));
    assert(prog && "Out of mem");

    token_t* tok = NULL;
    while ((tok=mcl_da_token_pop_front(tokens)) != NULL) {
        switch(tok->type) {
            case TT_IDENT: {
                // printf("IDENT: %s\n", tok->text);
                if (!expect_token_err(tokens, (token_t){.type=TT_COLON })) return NULL;
                if (!expect_token_err(tokens, (token_t){.type=TT_COLON })) return NULL;
                token_t* type = mcl_da_token_pop_front(tokens);

                if (type->type == TT_KW_FN) {
                    ast_ident_t name = {
                        .name = tok->text,
                        .loc = tok->loc
                    };
                    mcl_da_ast_node_push(&prog->body, parse_func(tokens, name));
                } else if (type->type == TT_KW_STRUCT) {

                } else if (type->type == TT_KW_ENUM) {

                }
                return NULL;
                
            } break;
        }
        // mcl_da_ast_node_push(prog->body, );
    }
    return prog;
}


ast_func_decl_t* parse_func(mcl_da_token_t* tokens, ast_ident_t name) {
    ast_func_decl_t* func = malloc(sizeof(ast_func_decl_t)); 
    // parse args
    if (!expect_token_err(tokens, (token_t){.type=TT_PAREN_L})) return NULL;
    while (!try_token(tokens, (token_t){.type=TT_PAREN_R})) {

        // Name
        token_t* arg_name = expect_token_err(tokens,(token_t){.type=TT_IDENT});
        if (!arg_name) return NULL;
        
        if (!expect_token_err(tokens, (token_t){.type=TT_COLON})) return NULL;

        // Type
        token_t* type = expect_token_err(tokens, (token_t){.type=TT_IDENT});
        if (!type) return NULL;

        ast_func_param_t param = {
            .name = { .name = arg_name->text, .loc = arg_name->loc },
            .type = { .name = { .name = type->text, .loc = type->loc}, .inner=NULL},
        };
        mcl_da_func_param_push(&func->params, param);
        
        if (!try_token(tokens, (token_t){.type=TT_COMMA})) {
            if (!expect_token_err(tokens, (token_t){.type=TT_PAREN_R})) return NULL;
        }
    }
    // parse body
    return func;
}

token_t* expect_token(mcl_da_token_t* tks, token_t t) {
    token_t* _tok_ = mcl_da_token_pop_front(tks);
    if (_tok_->type != t.type) {
        return NULL;
    }
    return _tok_;
}
token_t* try_token(mcl_da_token_t* tks, token_t t) {
    token_t* _tok_ = mcl_da_token_peek_front(tks);
    if (_tok_->type != t.type) {
        return NULL;
    }
    return _tok_;
}
token_t* expect_token_err(mcl_da_token_t* tks, token_t t) {
    token_t* _tok_ = mcl_da_token_pop_front(tks);
    if (_tok_->type != t.type) {
        mcl_log(ERROR, "Expected %s", token_to_string(&t));
        mcl_log_loc(INFO, &_tok_->loc, "Got %s", token_to_string(_tok_));
        return NULL;

    }
    return _tok_;
}

