#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "token.h"

char* escape_str(char* str);

char* token_to_string(token_t* t) {
    char* buf = {0};
    int len = 0;
    switch(t->type) {
        case TT_IDENT:
            // printf("IDENT\n");
            len = strlen(t->text)+10;
            buf = malloc(len);
            snprintf(buf, len, "IDENT(\"%s\")", escape_str(t->text));
            return buf;
        case TT_STR: 
            // printf("STR\n");
            len = strlen(t->text)+10;
            buf = malloc(len);
            snprintf(buf, len, "STR(\"%s\")", t->text);
            buf = escape_str(t->text);
            return buf;
        case TT_CHR:
            // printf("CHR\n");
            buf = malloc(10);
            snprintf(buf, 10, "CHAR('%c')", *t->text);
            return buf;
        case TT_CURLY_R:
            return "}";
        case TT_CURLY_L:
            return "{";
        case TT_BRACK_R:
            return "]";
        case TT_BRACK_L:
            return "[";
        case TT_PAREN_R:
            return ")";
        case TT_PAREN_L:
            return "(";
        case TT_COLON:
            return ":";
        case TT_SEMI:
            return ";";
        case TT_COMMA:
            return ",";
        case TT_DOT:
            return ".";
        case TT_AMP:
            return "&";
        case TT_STAR:
            return "*";
        case TT_PLUS:
            return "+";
        case TT_DASH:
            return "-";
        case TT_FSLASH:
            return "/";
        case TT_BAR:
            return "|";
        case TT_EQ:
            return "=";
        case TT_LT:
            return "<";
        case TT_GT:
            return ">";
        case TT_KW_FN: 
            return "KEYWORD(fn)";
        case TT_KW_RETURN: 
            return "KEYWORD(return)";
        case TT_KW_IF: 
            return "KEYWORD(if)";
        case TT_KW_ELSE: 
            return "KEYWORD(else)";
        case TT_KW_FOR: 
            return "KEYWORD(for)";
        case TT_KW_WHILE: 
            return "KEYWORD(while)";
        case TT_KW_STRUCT: 
            return "KEYWORD(struct)";
        case TT_KW_ENUM: 
            return "KEYWORD(enum)";
        case TT_KW_BREAK: 
            return "KEYWORD(break)";
        case TT_KW_CONTINUE: 
            return "KEYWORD(continue)";
    }
}

char* escape_str(char* str) {
    // First, determine the length of the new string
    size_t new_length = 0;
    for (const char* p = str; *p != '\0'; ++p) {
        if (*p == '\n' || *p == '\r') {
            new_length += 2; // Each '\n' or '\r' becomes two characters: '\\' and 'n' or 'r'
        } else {
            new_length += 1;
        }
    }

    // Allocate memory for the new string
    char* new_str = (char*)malloc(new_length + 1); // +1 for the null terminator
    if (new_str == NULL) {
        fprintf(stderr, "Memory allocation failed\n");
        exit(EXIT_FAILURE);
    }

    // Replace '\n' with "\\n" and '\r' with "\\r"
    char* q = new_str;
    for (const char* p = str; *p != '\0'; ++p) {
        if (*p == '\n') {
            *q++ = '\\';
            *q++ = 'n';
        } else if (*p == '\r') {
            *q++ = '\\';
            *q++ = 'r';
        } else {
            *q++ = *p;
        }
    }
    *q = '\0'; // Null-terminate the new string

    return new_str;
}